import logging
from collections import namedtuple
from typing import Union
from enum import Enum, unique

from ._daemon import sd_notify


log = logging.getLogger("cysystemd.daemon")
NotificationValue = namedtuple(
    "NotificationValue", ("name", "constant", "type")
)


@unique
class Notification(Enum):
    READY = NotificationValue(name="READY", constant=1, type=int)
    RELOADING = NotificationValue(name="RELOADING", constant=1, type=int)
    STOPPING = NotificationValue(name="STOPPING", constant=1, type=int)
    STATUS = NotificationValue(name="STATUS", constant=None, type=str)
    ERRNO = NotificationValue(name="ERRNO", constant=None, type=int)
    BUSERROR = NotificationValue(name="BUSERROR", constant=None, type=str)
    MAINPID = NotificationValue(name="MAINPID", constant=None, type=int)
    WATCHDOG = NotificationValue(name="WATCHDOG", constant=1, type=int)
    FDSTORE = NotificationValue(name="FDSTORE", constant=1, type=int)
    FDNAME = NotificationValue(name="FDNAME", constant=None, type=int)
    WATCHDOG_USEC = NotificationValue(
        name="WATCHDOG_USEC", constant=None, type=int
    )


def notify(
    notification: Notification,
    value: Union[str, int] = None,
    unset_environment: bool = False,
    return_exceptions: bool = True,
):

    """ Send notification to systemd daemon

    :param return_exceptions: Return exception or raise it.
    :param unset_environment: If the unset_environment parameter is non-zero,
        notify() will unset the $NOTIFY_SOCKET environment variable
        before returning (regardless of whether the function call itself
        succeeded or not). Further calls to notify() will then fail,
        but the variable is no longer inherited by child processes.
    :param notification: Notification object
    :param value: str or int value for non constant notifications
    :returns None
    """

    if not isinstance(notification, Notification):
        raise TypeError("state must be an instance of Notification")

    state = notification.value

    if state.constant is not None and value:
        raise ValueError(
            "State %s should contain only constant value %r"
            % (state.name, state.constant),
            state.name,
            state.constant,
        )

    line = "%s=%s" % (
        state.name,
        state.constant if state.constant is not None else state.type(value),
    )

    log.debug("Send %r into systemd", line)

    try:
        return sd_notify(line, unset_environment)
    except Exception as e:
        if return_exceptions:
            log.error("%s", e)
            return e
        raise


__all__ = ("notify", "Notification")



import logging
import re
import traceback
import uuid
from enum import IntEnum, unique

from ._journal import send, syslog_priorities


try:
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping


_priorities = syslog_priorities()


__all__ = "write", "send", "Priority", "JournaldLogHandler", "Facility"


@unique
class Priority(IntEnum):
    PANIC = _priorities["panic"]
    WARNING = _priorities["warn"]
    ALERT = _priorities["alert"]
    NONE = _priorities["none"]
    CRITICAL = _priorities["crit"]
    DEBUG = _priorities["debug"]
    INFO = _priorities["info"]
    ERROR = _priorities["error"]
    NOTICE = _priorities["notice"]


@unique
class Facility(IntEnum):
    KERN = 0
    USER = 1
    MAIL = 2
    DAEMON = 3
    AUTH = 4
    SYSLOG = 5
    LPR = 6
    NEWS = 7
    UUCP = 8
    CLOCK_DAEMON = 9
    AUTHPRIV = 10
    FTP = 11
    NTP = 12
    AUDIT = 13
    ALERT = 14
    CRON = 15
    LOCAL0 = 16
    LOCAL1 = 17
    LOCAL2 = 18
    LOCAL3 = 19
    LOCAL4 = 20
    LOCAL5 = 21
    LOCAL6 = 22
    LOCAL7 = 23


def write(message, priority=Priority.INFO):
    """ Write message into systemd journal
    :type priority: Priority
    :type message: str
    """

    priority = int(Priority(int(priority)))

    send(priority=priority, message=message)


class JournaldLogHandler(logging.Handler):
    FIELD_BADCHAR_RE = re.compile(r'\W')
    LEVELS = {
        logging.CRITICAL: Priority.CRITICAL.value,
        logging.FATAL: Priority.PANIC.value,
        logging.ERROR: Priority.ERROR.value,
        logging.WARNING: Priority.WARNING.value,
        logging.WARN: Priority.WARNING.value,
        logging.INFO: Priority.INFO.value,
        logging.DEBUG: Priority.DEBUG.value,
        logging.NOTSET: Priority.NONE.value,
    }

    __slots__ = ("__facility",)

    def __init__(self, identifier=None, facility=Facility.DAEMON):
        """

        :type identifier: Override default journald identifier
        :type facility: Facility
        """
        logging.Handler.__init__(self)
        self.__identifier = identifier
        self.__facility = int(facility)

    @staticmethod
    def _to_microsecond(ts):
        """

        :type ts: float
        """
        return int(ts * 1000 * 1000)

    def emit(self, record):
        message = str(record.getMessage())

        tb_message = ""
        if record.exc_info:
            tb_message = "\n".join(
                traceback.format_exception(*record.exc_info)
            )

        message += "\n"
        message += tb_message

        ts = self._to_microsecond(record.created)

        hash_fields = (
            message,
            record.funcName,
            record.levelno,
            record.process,
            record.processName,
            record.levelname,
            record.pathname,
            record.name,
            record.thread,
            record.lineno,
            ts,
            tb_message,
        )

        message_id = uuid.uuid3(
            uuid.NAMESPACE_OID, "$".join(str(x) for x in hash_fields)
        ).hex

        data = {
            key: value
            for key, value in record.__dict__.items()
            if not key.startswith("_") and value is not None
        }
        data["message"] = self.format(record)
        data["priority"] = self.LEVELS[data.pop("levelno")]
        data["syslog_facility"] = self.__facility

        data["code_file"] = data.pop("filename")
        data["code_line"] = data.pop("lineno")
        data["code_func"] = data.pop("funcName")
        if self.__identifier:
            data["syslog_identifier"] = self.__identifier
        else:
            data["syslog_identifier"] = data["name"]

        if "msg" in data:
            data["message_raw"] = data.pop("msg")

        data["message_id"] = message_id
        data["code_module"] = data.pop("module")
        data["logger_name"] = data.pop("name")
        data["pid"] = data.pop("process")
        data["proccess_name"] = data.pop("processName")
        data["errno"] = 0 if not record.exc_info else 255
        data["relative_ts"] = self._to_microsecond(data.pop("relativeCreated"))
        data["thread_name"] = data.pop("threadName")

        args = data.pop("args", [])
        if isinstance(args, Mapping):
            for key, value in args.items():
                key = self.FIELD_BADCHAR_RE.sub('_', key)
                data["argument_%s" % key] = value
        else:
            for idx, item in enumerate(args):
                data["argument_%d" % idx] = str(item)

        if tb_message:
            data["traceback"] = tb_message

        send(**data)


handler = JournaldLogHandler()


class JournaldLogger(logging.Logger):
    def __init__(self, level, name="root"):
        super(JournaldLogger, self).__init__(name, level)
        self.addHandler(handler)


Logger = JournaldLogger(logging.WARNING)



package_info = "systemd wrapper in Cython"
version_info = (1, 6, 3)


author_info = (("Dmitry Orlov", "me@mosquito.su"),)

author_email = ", ".join("{}".format(info[1]) for info in author_info)

license = "Apache 2.0"

__version__ = ".".join(str(x) for x in version_info)
__author__ = ", ".join("{} <{}>".format(*info) for info in author_info)


__all__ = (
    "__author__",
    "__version__",
    "author_info",
    "license",
    "package_info",
    "version_info",
)



import asyncio
import logging
import threading
from collections import deque

from collections.abc import AsyncIterator
from functools import partial
from typing import Callable, TypeVar
from uuid import UUID
from weakref import finalize

from .reader import JournalOpenMode, JournalReader, JournalEntry


A = TypeVar("A")
R = TypeVar("R")
log = logging.getLogger("cysystemd.async_reader")


class Base:
    def __init__(self, loop=None, executor=None):
        self._executor = executor
        self._loop = loop or asyncio.get_event_loop()

    async def _exec(self, func: Callable[[A], R], *args, **kwargs) -> R:
        # noinspection PyTypeChecker
        return await self._loop.run_in_executor(
            self._executor, partial(func, *args, **kwargs)
        )


class AsyncJournalReader(Base):
    def __init__(self, executor=None, loop=None):
        super().__init__(loop=loop, executor=executor)
        self.__reader = JournalReader()
        self.__flags = None
        self.__wait_lock = asyncio.Lock()
        self.__iterator = None

    async def wait(self):
        async with self.__wait_lock:
            loop = self._loop
            reader = self.__reader
            event = asyncio.Event()

            loop.add_reader(reader.fd, event.set)

            try:
                await event.wait()
            finally:
                loop.remove_reader(reader.fd)

            reader.process_events()

        return True

    def open(self, flags=JournalOpenMode.CURRENT_USER):
        self.__flags = flags
        return self._exec(self.__reader.open, flags=flags)

    def open_directory(self, path):
        return self._exec(self.__reader.open_directory, path)

    def open_files(self, *file_names):
        return self._exec(self.__reader.open_files, *file_names)

    @property
    def data_threshold(self):
        return self.__reader.data_threshold

    @data_threshold.setter
    def data_threshold(self, size):
        self.__reader.data_threshold = size

    @property
    def closed(self):
        return self.__reader.closed

    @property
    def locked(self):
        return self.__reader.locked

    @property
    def idle(self):
        return self.__reader.idle

    def seek_head(self):
        return self._exec(self.__reader.seek_head)

    def __repr__(self):
        return "<%s[%s]: %s>" % (
            self.__class__.__name__,
            self.__flags,
            "closed" if self.closed else "opened",
        )

    @property
    def fd(self):
        return self.__reader.fd

    @property
    def events(self):
        return self.__reader.events

    @property
    def timeout(self):
        return self.__reader.timeout

    def get_catalog(self):
        return self._exec(self.__reader.get_catalog)

    def get_catalog_for_message_id(self, message_id):
        return self._exec(
            self.__reader.get_catalog_for_message_id, message_id
        )

    def seek_tail(self):
        return self._exec(self.__reader.seek_tail)

    def seek_monotonic_usec(self, boot_id: UUID, usec):
        return self._exec(
            self.__reader.seek_monotonic_usec, boot_id, usec
        )

    def seek_realtime_usec(self, usec):
        return self._exec(self.__reader.seek_realtime_usec, usec)

    def seek_cursor(self, cursor):
        return self._exec(self.__reader.seek_cursor, cursor)

    def skip_next(self, skip):
        return self._exec(self.__reader.skip_next, skip)

    def previous(self, skip=0):
        return self._exec(self.__reader.previous, skip)

    def skip_previous(self, skip):
        return self._exec(self.__reader.skip_previous, skip)

    def add_filter(self, rule):
        return self._exec(self.__reader.add_filter, rule)

    def clear_filter(self):
        return self._exec(self.__reader.clear_filter)

    def next(self, skip=0):
        return self._exec(self.__reader.next, skip)

    def __aiter__(self) -> "AsyncReaderIterator":
        if self.__iterator is not None:
            self.__iterator.close()
            self.__iterator = None

        iterator = AsyncReaderIterator(
            loop=self._loop, executor=self._executor, reader=self.__reader
        )

        finalize(self, iterator.close)

        self.__iterator = iterator
        return iterator


class AsyncReaderIterator(Base, AsyncIterator):
    __slots__ = "reader", "queue", "queue_full", "event", "lock", "closed"

    QUEUE_SIZE = 2
    WRITE_EVENT_WAIT_TIME = 0.1

    def __init__(self, *, reader, loop, executor):
        super().__init__(loop=loop, executor=executor)
        self.reader = reader
        self.lock = asyncio.Lock()
        self.queue = deque()
        self.read_event = asyncio.Event()
        self.write_event = threading.Semaphore(self.QUEUE_SIZE)
        self.close_event = threading.Event()

        self._loop.create_task(self._exec(self._journal_reader))

    def close(self):
        self.close_event.set()
        self.__set_read_event()

    def __del__(self):
        self.close()

    def __set_read_event(self):
        if self._loop.is_closed():
            return

        self._loop.call_soon_threadsafe(self.read_event.set)

    def _journal_reader(self):
        try:
            for item in self.reader:
                while not self.close_event.is_set():
                    if self.write_event.acquire(
                        timeout=self.WRITE_EVENT_WAIT_TIME
                    ):
                        break
                else:
                    return

                self.queue.append(item)
                self.__set_read_event()
        finally:
            self.close()

    async def __anext__(self) -> JournalEntry:
        async with self.lock:
            if self.close_event.is_set() and len(self.queue) == 0:
                raise StopAsyncIteration

            while True:
                try:
                    item = self.queue.popleft()
                except IndexError:
                    await self.read_event.wait()
                    self.read_event.clear()
                    continue
                else:
                    self.write_event.release()

                return item
